#include "eeprom.h"

#define SLAVE_I2C_GENERIC_RETRY_MAX 100

void EMULATED_EEPROM_Write(
                                uint16_t slaveDeviceAddress,
                                uint16_t dataAddress,
                                uint8_t *pData,
                                uint16_t nCount)
{
    // write to an EEPROM Device

    //uint8_t         writeBuffer[3];
    uint8_t         writeBuffer[2];
    uint8_t         *pD;
    uint16_t        counter, timeOut;

    I2C_MESSAGE_STATUS status = I2C_MESSAGE_PENDING;
/*
    dataAddress = 0x10;             // starting EEPROM address 
    pD = sourceData;                // initialize the source of the data
    nCount = 16;                    // number of bytes to write
*/
    pD = pData;
    
    for (counter = 0; counter < nCount; counter++)
    {
        //printf("here0(%02X)->", dataAddress);
        // build the write buffer first
        // starting address of the EEPROM memory
        //writeBuffer[0] = (dataAddress >> 8);                // high address
        //writeBuffer[1] = (uint8_t)(dataAddress);            // low low address
        writeBuffer[0] = (uint8_t)(dataAddress);            // low low address

        //printf("(%02X),", *pD);
        // data to be written
        //writeBuffer[2] = *pD++;
        writeBuffer[1] = *pD++;

        // Now it is possible that the slave device will be slow.
        // As a work around on these slaves, the application can
        // retry sending the transaction
        timeOut = 0;
        while(status != I2C_MESSAGE_FAIL)
        {
            //printf("here1,");
            // write one byte to EEPROM (3 is the number of bytes to write)
            I2C_MasterWrite(  writeBuffer,
                                    //3,
                                    2,
                                    slaveDeviceAddress,
                                    &status);

            // wait for the message to be sent or status has changed.
            while(status == I2C_MESSAGE_PENDING) {
                __delay_ms(5);
            }

            //printf("here2(%d),", status);
            if (status == I2C_MESSAGE_COMPLETE)
                break;

            // if status is  I2C_MESSAGE_ADDRESS_NO_ACK,
            //               or I2C_DATA_NO_ACK,
            // The device may be busy and needs more time for the last
            // write so we can retry writing the data, this is why we
            // use a while loop here

            //printf("here3(%d),", status);
            // check for max retry and skip this byte
            if (timeOut == SLAVE_I2C_GENERIC_RETRY_MAX)
                break;
            else
                timeOut++;
        }

        if (status == I2C_MESSAGE_FAIL)
        {
            break;
        }
        dataAddress++;

    }
    //printf("\r\n");
}

void EMULATED_EEPROM_Read(
                               uint16_t slaveDeviceAddress,
                               uint16_t dataAddress,
                               uint8_t *pData,
                               uint16_t nCount)
{
    I2C_MESSAGE_STATUS status;
    I2C_TRANSACTION_REQUEST_BLOCK readTRB[2];
    uint8_t     writeBuffer[3];
    uint16_t    timeOut;

    // this initial value is important
    status = I2C_MESSAGE_PENDING;

    // build the write buffer first
    // starting address of the EEPROM memory
    //writeBuffer[0] = (dataAddress >> 8);                        // high address
    //writeBuffer[1] = (uint8_t)(dataAddress);                    // low low address
    writeBuffer[0] = (uint8_t)(dataAddress);                    // low low address

    // we need to create the TRBs for a random read sequence to the EEPROM
    // Build TRB for sending address
    I2C_MasterWriteTRBBuild(   &readTRB[0],
                                    writeBuffer,
                                    //2,
                                    1,
                                    slaveDeviceAddress);
    // Build TRB for receiving data
    I2C_MasterReadTRBBuild(    &readTRB[1],
                                    pData,
                                    nCount,
                                    slaveDeviceAddress);

    timeOut = 0;

    while(status != I2C_MESSAGE_FAIL)
    {
        // now send the transactions
        I2C_MasterTRBInsert(2, readTRB, &status);

        // wait for the message to be sent or status has changed.
        while(status == I2C_MESSAGE_PENDING);

        if (status == I2C_MESSAGE_COMPLETE)
            break;

        // if status is  I2C_MESSAGE_ADDRESS_NO_ACK,
        //               or I2C_DATA_NO_ACK,
        // The device may be busy and needs more time for the last
        // write so we can retry writing the data, this is why we
        // use a while loop here

        // check for max retry and skip this byte
        if (timeOut == SLAVE_I2C_GENERIC_RETRY_MAX)
            break;
        else
            timeOut++;

    }
}   
